//	Distributed under the Boost Software License Version 1.0 https://www.boost.org/LICENSE_1_0.txt
//	Copyright Gero Peterhoff

#ifndef BOOST_QUADMATH_GCC_LIMITS_HPP
#define BOOST_QUADMATH_GCC_LIMITS_HPP

#include <array>
#include <limits>
#include <boost/config.hpp>
#include <boost/quadmath/quadmath_gcc.hpp>
#include <boost/type_traits/endian.hpp>

#if defined(__cpp_lib_bit_cast)
#include <bit>
#endif



#if BOOST_GCC
namespace std
{
#if defined(__STDCPP_FLOAT128_T__)
inline constexpr __float128	nansq(const char*const)	noexcept
{
	return __float128(numeric_limits<std::float128_t>::signaling_NaN());
}
#elif defined(__cpp_lib_bit_cast)
inline constexpr __float128	nansq(const char*const)	noexcept
{
	using array_t = std::array<uint64_t, 2>;

	BOOST_IF_CONSTEXPR (boost::is_little_endian::value)	return bit_cast<__float128>(array_t{0ULL, 9223160930622242816ULL});
	else												return bit_cast<__float128>(array_t{9223160930622242816ULL, 0ULL});
}
#else
inline __float128	nansq(const char*const)	noexcept
{
	using array_t = std::array<uint64_t, 2>;

	BOOST_IF_CONSTEXPR (boost::is_little_endian::value)	return reinterpret_cast<__float128>(array_t{0ULL, 9223160930622242816ULL});
	else												return reinterpret_cast<__float128>(array_t{9223160930622242816ULL, 0ULL});
}
#endif



template <> struct numeric_limits<__float128>
{
static constexpr bool				is_specialized		= true;
static constexpr __float128			min()				noexcept	{ return FLT128_MIN;	}
static constexpr __float128			max()				noexcept	{ return FLT128_MAX;	}
static constexpr __float128			lowest()			noexcept	{ return -FLT128_MAX;	}
static constexpr int				digits				= FLT128_MANT_DIG;
static constexpr int				digits10			= FLT128_DIG;
static constexpr int				max_digits10		= 36;
static constexpr bool				is_signed			= true;
static constexpr bool				is_integer			= false;
static constexpr bool				is_exact			= false;
static constexpr int				radix				= 2;
static constexpr __float128			epsilon()			noexcept	{ return FLT128_EPSILON;}
static constexpr __float128			round_error()		noexcept	{ return 0.5Q;			}
static constexpr int				min_exponent		= FLT128_MIN_EXP;
static constexpr int				min_exponent10		= FLT128_MIN_10_EXP;
static constexpr int				max_exponent		= FLT128_MAX_EXP;
static constexpr int				max_exponent10		= FLT128_MAX_10_EXP;
static constexpr bool				has_infinity		= true;
static constexpr bool				has_quiet_NaN		= true;
static constexpr bool				has_signaling_NaN	= bool(issignalingq(0.0Q));
static constexpr float_denorm_style	has_denorm			= denorm_present;
static constexpr bool				has_denorm_loss		= false;
static constexpr __float128			infinity()			noexcept	{ return HUGE_VALQ;	}
static constexpr __float128			quiet_NaN()			noexcept	{ return nanq("");	}

#if defined(__STDCPP_FLOAT128_T__) || defined(__cpp_lib_bit_cast)
static constexpr __float128			signaling_NaN()		noexcept	{ return nansq("");	}
#else
static inline	 __float128			signaling_NaN()		noexcept	{ return nansq("");	}
#endif

static constexpr __float128			denorm_min()		noexcept	{ return FLT128_DENORM_MIN;	}
static constexpr bool				is_iec559			= true;
static constexpr bool				is_bounded			= true;
static constexpr bool				is_modulo			= false;
static constexpr bool				traps				= false;
static constexpr bool				tinyness_before		= false;
static constexpr float_round_style	round_style			= round_to_nearest;
};
}	//	std
#endif	//	BOOST_GCC
#endif	//	BOOST_QUADMATH_GCC_LIMITS_HPP
